/****************************************************************************
*
* create_ObjDesc
*
* History : the initial iPSC/860 version on Sep/15/93
*	    the second revision for scatter_back on Oct/07/93
* Author  : Bongki Moon (bkmoon@cs.umd.edu)
*
*****************************************************************************/

#include <stdio.h>
#include <varargs.h>
#include <assert.h>
#include "parti.h"

typedef struct object_type {
	char		*saddr;		/* starting address */
	int		size;		/* size of individual element */
	int		offset;		/* element offset in bytes */
} objtype;
typedef struct object_desc {
	int		nobjs;		/* number of objects */
	int		size;		/* object size in bytes */
	objtype		*object;	/* pointer to a list of objects */
} ObjDesc;
typedef int		DirType;	/* direction of data migration */
typedef int	rint;
typedef char	rchar;

#define		APND		1
#define		BACK		2

/*#define COMTIME 1*/
#ifdef COMTIME
extern double	dclock();
static double	tcomm=0., ttmp;
print_tcomm_()
{
	double	work;

	gdhigh(&tcomm,1,&work);
	if (PARTI_mynode==0) printf("Max. Tcomm = %f\n",tcomm);
}
#endif

/* create_ObjDesc(n_objs,{object,size}) */
ObjDesc*
create_ObjDesc(va_alist)
va_dcl
{
	va_list		pvar;
	ObjDesc		*desc, *BuildObjectDesc();
	char		**saddr;
	int		*sizes, nobjs, i;

	va_start(pvar);
	nobjs = va_arg(pvar,int);
	saddr = (char**)malloc(nobjs * sizeof(char*));
	sizes = (int*)malloc(nobjs * sizeof(int));
	for(i=0; i < nobjs ;i++) {
		saddr[i] = va_arg(pvar,char*);
		sizes[i] = va_arg(pvar,int);
	}
	va_end(pvar);

	desc = BuildObjectDesc(nobjs,saddr,sizes);

	free(saddr); free(sizes);
	return(desc);
}

ObjDesc*
BuildObjectDesc(nobjs,saddr,sizes)
int	nobjs;
char	**saddr;
int	*sizes;
{
	ObjDesc		*desc;
	objtype		*objp;
	int		i;

	desc = (ObjDesc*)malloc(sizeof(ObjDesc));
	desc->nobjs = nobjs;
	desc->object = (objtype*)malloc(desc->nobjs * sizeof(objtype));

	objp = desc->object;
	for(i=0; i < desc->nobjs ;i++) {
		objp[i].saddr = saddr[i];
		objp[i].size = sizes[i];
	}

	objp[0].offset = 0;
	for(i=1; i < desc->nobjs ;i++)
		objp[i].offset = objp[i-1].offset + objp[i-1].size;
	desc->size = objp[desc->nobjs-1].offset + objp[desc->nobjs-1].size;

	return(desc);
}

addObjectDesc(desc,saddr,size)
ObjDesc	*desc;
char	*saddr;
int	size;
{
	objtype		*oldp, *objp;
	int		i;

	oldp = desc->object;
	objp = (objtype*)malloc((desc->nobjs+1) * sizeof(objtype));
	for(i=0; i < desc->nobjs ;i++) {
		objp[i].saddr = oldp[i].saddr;
		objp[i].size = oldp[i].size;
		objp[i].offset = oldp[i].offset;
	}
	free(oldp);
	objp[desc->nobjs].saddr = saddr;
	objp[desc->nobjs].size = size;
	objp[desc->nobjs].offset = objp[desc->nobjs-1].offset
					+ objp[desc->nobjs-1].size;
	desc->nobjs++;
	desc->size += size;
	desc->object = objp;
}

free_desc(desc)
ObjDesc	*desc;
{
	free(desc->object);
	free(desc);
}

/* this alpha version contains FORWARD only */
ObjMigrate(sched,desc,direction)
SCHED	*sched;
ObjDesc	*desc;
DirType	direction;
{
	int	i, j, k, bsize, wsize, bcnt, wcnt, dstoff;
	char	**blist, **wlist, *sbuf, *rbuf;
	int	*permp, *self_permp;
	void	(*cpfunc)(), execute_migration();
	void	wdbuf(), bybuf(), wdunbuf(), byunbuf(), wdshift(), byshift();

	blist = wbuffer->blist;
	wlist = wbuffer->wlist;
	bsize = sched->sr->total_fetch_size * desc->size;
	wsize = sched->sr->total_send_size * desc->size;
	sbuf = (char*)malloc(bsize);
	rbuf = (char*)malloc(wsize);

	for(bcnt=wcnt=0,i=0; i<PARTI_numnodes ;i++){
		if (i == PARTI_mynode) continue;
		blist[i] = &sbuf[bcnt];
		bcnt += sched->sr->fetch_n[i] * desc->size;
		wlist[i] = &rbuf[wcnt];
		wcnt += sched->sr->send_n[i] * desc->size;
	}

	permp = sched->perm;
	for(i=0; i < PARTI_numnodes ;i++) {
		if (sched->sr->fetch_n[i]==0) continue;
		if (i==PARTI_mynode) {
			self_permp = permp;
			permp += sched->sr->fetch_n[PARTI_mynode];
			continue;
		}
		for(j=0; j < desc->nobjs ;j++) {
			cpfunc = (desc->object[j].size%sizeof(int)==0)
					? wdbuf : bybuf;
			cpfunc(desc,j,blist[i],permp,sched->sr->fetch_n[i]);
		}
		permp += sched->sr->fetch_n[i];
	}

	for(j=0; j < desc->nobjs ;j++) {
		cpfunc = (desc->object[j].size%sizeof(int)==0)
				? wdshift : byshift;
		cpfunc(desc,j,self_permp,sched->sr->fetch_n[PARTI_mynode]);
	}

#ifdef COMTIME
ttmp = dclock();
	execute_migration(sched->sr,desc,blist,wlist,direction);
tcomm += dclock() - ttmp;
#else
	execute_migration(sched->sr,desc,blist,wlist,direction);
#endif

	dstoff = sched->sr->fetch_n[PARTI_mynode];
	for(i=0; i < PARTI_numnodes ;i++) {
		if (i == PARTI_mynode) continue;
		if (sched->sr->send_n[i]==0) continue;
		for(j=0; j < desc->nobjs ;j++) {
			cpfunc = (desc->object[j].size%sizeof(int)==0)
					? wdunbuf : byunbuf;
			cpfunc(desc,j,wlist[i],dstoff,sched->sr->send_n[i]);
		}
		dstoff += sched->sr->send_n[i];
	}

	free(sbuf); free(rbuf);
}

void
wdbuf(desc,oid,sbufp,permp,ndata)
ObjDesc	*desc;
int	oid;
char	*sbufp;
int	*permp;
int	ndata;
{
	int	offset = desc->object[oid].offset;
	int	nbytes = desc->object[oid].size;
	int	nwords = nbytes / sizeof(int);
	rint	*srcp = (int*)desc->object[oid].saddr;
	rint	*dstp = (int*)(sbufp + offset);
	int	i, j;

	for(i=0; i < ndata ;i++,permp++) {
		for(j=0; j < nwords ;j++) *dstp++ = srcp[*permp*nwords + j];
		dstp = (int*)((char*)dstp + (desc->size - nbytes));
	}
}

void
bybuf(desc,oid,sbufp,permp,ndata)
ObjDesc	*desc;
int	oid;
char	*sbufp;
int	*permp;
int	ndata;
{
	int	offset = desc->object[oid].offset;
	int	nbytes = desc->object[oid].size;
	rchar	*srcp = desc->object[oid].saddr;
	rchar	*dstp = sbufp + offset;
	int	i, j;

	for(i=0; i < ndata ;i++,permp++) {
		for(j=0; j < nbytes ;j++) *dstp++ = srcp[*permp + j];
		dstp = dstp + (desc->size - nbytes);
	}
}

void
wdunbuf(desc,oid,rbufp,dstoff,ndata)
ObjDesc	*desc;
int	oid;
char	*rbufp;
int	dstoff;
int	ndata;
{
	int	offset = desc->object[oid].offset;
	int	nbytes = desc->object[oid].size;
	int	nwords = nbytes / sizeof(int);
	rint	*dstp = (int*)(desc->object[oid].saddr + dstoff*nbytes);
	rint	*srcp = (int*)(rbufp + offset);
	int	i, j;

	for(i=0; i < ndata ;i++) {
		for(j=0; j < nwords ;j++) *dstp++ = srcp[j];
		srcp = (int*)((char*)srcp + desc->size);
	}
}

void
byunbuf(desc,oid,rbufp,dstoff,ndata)
ObjDesc	*desc;
int	oid;
char	*rbufp;
int	dstoff;
int	ndata;
{
	int	offset = desc->object[oid].offset;
	int	nbytes = desc->object[oid].size;
	rchar	*dstp = desc->object[oid].saddr + dstoff*nbytes;
	rchar	*srcp = rbufp + offset;
	int	i, j;

	for(i=0; i < ndata ;i++) {
		for(j=0; j < nbytes ;j++) *dstp++ = srcp[j];
		srcp = srcp + desc->size;
	}
}

void
wdshift(desc,oid,permp,ndata)
ObjDesc	*desc;
int	oid;
int	*permp;
int	ndata;
{
	int	nbytes = desc->object[oid].size;
	int	nwords = nbytes / sizeof(int);
	rint	*srcp = (int*)desc->object[oid].saddr;
	rint	*dstp = srcp;
	int	i, j;

	for(i=0; i < ndata ;i++,permp++)
		for(j=0; j < nwords ;j++) *dstp++ = srcp[*permp*nwords + j];
}

void
byshift(desc,oid,permp,ndata)
ObjDesc	*desc;
int	oid;
int	*permp;
int	ndata;
{
	int	nbytes = desc->object[oid].size;
	rint	*srcp = (int*)desc->object[oid].saddr;
	rint	*dstp = srcp;
	int	i, j;

	for(i=0; i < ndata ;i++,permp++)
		for(j=0; j < nbytes ;j++) *dstp++ = srcp[*permp + j];
}

/* ASYNCHRONOUS COMMUNICATION */
void
execute_migration(pattern,desc,blist,wlist,dir)
struct com_pattern *pattern;
ObjDesc	*desc;
char	**blist, **wlist;
int	dir;
{
	static int call_count=0;
	int	size = desc->size;
	int	i, j, p, count;
	int	*msg_id, *rcv_n, *snd_n, *sync_id;
	char	**src, **dst, sync_buf[1];

	/* message buffer */
	msg_id = wbuffer->msg_id;
	sync_id = (int*)malloc(sizeof(int) * PARTI_numnodes);

	/* forward or backward migration */
	if (dir == APND) {
		src = blist; dst = wlist;
		rcv_n = pattern->send_n; snd_n = pattern->fetch_n;
	} else {
		src = wlist; dst = blist;
		rcv_n = pattern->fetch_n; snd_n = pattern->send_n;
	}

	/* send and receive messages ASYNCHRONOUSLY -- Bongki Moon, Jan/03/94 */
	for(count=0,p=(PARTI_mynode+1)%PARTI_numnodes; p!=PARTI_mynode ;p=(p+1)%PARTI_numnodes) {
	    if (rcv_n[p]) {
		msg_id[count] = PARTI_irecv(FORCED_SCATTER_EXECUTOR_MSG+
			call_count*MAXNODES+p,dst[p],rcv_n[p]*size
			,p, PARTI_pid);
		sync_id[count++] = PARTI_isend(SCATTER_SYNC+call_count*MAXNODES+
			PARTI_mynode,sync_buf,0,p,PARTI_pid);
	    }
	}

	for(p=(PARTI_mynode-1+PARTI_numnodes)%PARTI_numnodes; p!=PARTI_mynode ;p=(p-1+PARTI_numnodes)%PARTI_numnodes) {
	    if (snd_n[p]) {
		PARTI_crecv(SCATTER_SYNC+call_count*MAXNODES+p,sync_buf,0,p,
			PARTI_pid);
		PARTI_csend(FORCED_SCATTER_EXECUTOR_MSG+call_count*MAXNODES+
			PARTI_mynode,src[p],size*snd_n[p],p,PARTI_pid);
	    }
	}

	/* wait messages and store data */
	for(count=0,p=(PARTI_mynode+1)%PARTI_numnodes; p!=PARTI_mynode ;p=(p+1)%PARTI_numnodes) {
		if (rcv_n[p]) {
			PARTI_msgwait(msg_id[count]);
			PARTI_msgwait(sync_id[count++]);
		}
	}

	free(sync_id);
	call_count = (++call_count > MAXCOUNT)? 0: call_count;
}

/* ----- wrapper ---------------------------------------------------------- */

/* create_ObjDesc(n_objs,{object,size}) */
ObjDesc*
create_objdesc_(va_alist)
va_dcl
{
	va_list		pvar;
	ObjDesc		*desc, *BuildObjectDesc();
	char		**saddr;
	int		*sizes, nobjs, i;

	va_start(pvar);
	nobjs = *va_arg(pvar,int*);
	saddr = (char**)malloc(nobjs * sizeof(char*));
	sizes = (int*)malloc(nobjs * sizeof(int));
	for(i=0; i < nobjs ;i++) {
		saddr[i] = va_arg(pvar,char*);
		sizes[i] = *va_arg(pvar,int*);
	}
	va_end(pvar);

	desc = BuildObjectDesc(nobjs,saddr,sizes);

	free(saddr); free(sizes);
	return(desc);
}

addobjectdesc_(desc,saddr,size)
ObjDesc **desc;
char    *saddr;
int     *size;
{
	addObjectDesc(*desc,saddr,*size);
}

free_desc_(desc)
ObjDesc **desc;
{
	free_desc(*desc);
}

objmigrate_(sched,desc,direction)
SCHED	**sched;
ObjDesc	**desc;
DirType	*direction;
{
	ObjMigrate(*sched,*desc,*direction);
}

print_desc_(fdesc)
ObjDesc	**fdesc;
{
	ObjDesc	*desc = *fdesc;
	int	i;

	printf("nobjs = %d\n",desc->nobjs);
	printf("obj size = %d\n",desc->size);
	for(i=0; i < desc->nobjs ;i++)
		printf("oid=%d, size=%d, offset=%d, saddr=0x%x\n",
		i,desc->object[i].size,desc->object[i].offset,
		desc->object[i].saddr);
}

/* end of migrate.c */
